/*
 * This driver is part of lrtbf. It log the time when it send a
 * signal and when it get a signal on the parallel port.
 * Copyright (C) 2005 Kristian Benoit, Opersys inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/relayfs_fs.h>
#include <linux/proc_fs.h>
#include <linux/adeos.h>

#define PAR_LOG_IRQ 7
#define PAR_LOG_NAME "par_log"
#define PAR_LOG_DEV_ID (void *)par_handler
#define PAR_LOG_BASE 0x378

#define PAR_LOG_PORT_DATA PAR_LOG_BASE
#define PAR_LOG_PORT_STATUS PAR_LOG_BASE+1
#define PAR_LOG_PORT_CONTROL PAR_LOG_BASE+2

#define PAR_LOG_CMD_ENABLE_INTR (1<<4)
#define PAR_LOG_CMD_DISABLE_INTR 0

#define BUFFER_SIZE (4096*1024) //4MB

cycles_t *buf;
static adomain_t domain;
static unsigned char state;
static adsysinfo_t sys_info;
static int read_proc_buf_pos = 0;
static unsigned long long nbintrx = 0;
static unsigned long long nbinttx = 0;
static int disable_timer = 0;
static int relay_id;

volatile static unsigned int read_pos = 0, write_pos = 0;

static inline int inc_read_pos (void)
{
	read_pos = (read_pos+1)%BUFFER_SIZE;
	return read_pos;
}

static inline int inc_write_pos (void)
{
	unsigned int newpos = (write_pos+1)%BUFFER_SIZE;
	if (read_pos == newpos)
		inc_read_pos ();
	write_pos = newpos;
	return write_pos;
}

#define MAX_LEN_OF_CYCLES_T 32 //The length of the string.
static int read_proc (char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	unsigned long flags;
	int len = 0;
	*start = page;
	*eof = 0;

	if (offset == 0)
		disable_timer = 1;

	flags = adeos_critical_enter (NULL);
	while (read_pos != write_pos && len <= (count-MAX_LEN_OF_CYCLES_T)) {
		len += sprintf ((*start)+len, "%llu\n", buf[read_pos]);
		inc_read_pos ();
	}
	adeos_critical_exit (flags);

	if (read_pos == write_pos) {
		disable_timer = 0;
		*eof = 1;
	}
	return len;
}

// From Ingo Molnar's lpptest (realtime-preempt-2.6.12-final-V0.7.50-11)
static cycles_t test_response(void)
{
	cycles_t now, end;
	unsigned char in;
	int timeout = 0;

//	local_irq_disable();
	in = inb(0x379);
	inb(0x378);
	outb(0xFF, 0x378);
	rdtscll(now);
	while(1) {
    		if (inb(0x379) != in)
			break;
		if (timeout++ > 1000000) {
			outb(0x00, 0x378);
			local_irq_enable();

			return 0;
		}
	}
	rdtscll(end);
	outb(0x00, 0x378);
//	local_irq_enable();

	return end - now;
}

static void irq_timer_handler (unsigned irq)
{
	unsigned long flags;

	if (!disable_timer) {
		nbinttx++;
		flags = adeos_critical_enter (NULL);
		buf[write_pos] = test_response();
		inc_write_pos();
		adeos_critical_exit (flags);
	}

	adeos_propagate_irq(irq);
}

static void domain_init (int iflag)
{
	printk ( "Domain %s started on CPU#%d (iflag=%d).\n",
			 adp_current->name,
			 adeos_processor_id(),
			 iflag);

	state = 0;
	outb (state, PAR_LOG_PORT_DATA);

	if (iflag)
	{
		adeos_get_sysinfo (&sys_info);

		adeos_virtualize_irq ( sys_info.archdep.tmirq,
							   &irq_timer_handler,
							   NULL,
							   IPIPE_DYNAMIC_MASK );

		outb (PAR_LOG_CMD_ENABLE_INTR, PAR_LOG_PORT_CONTROL);
	}

	while (42)
		adeos_suspend_domain();
}

static int __init initialize (void)
{
	static struct rchan_info chan;
	adattr_t attr;
	adeos_init_attr (&attr);
	attr.name		= "par_logger";
	attr.priority	= ADEOS_ROOT_PRI+2;
	attr.entry		= &domain_init;
	
	printk ("par_logger: Initializing ...\n");
	relay_id = relay_open ("/par_logger",
						   BUFFER_SIZE*sizeof(cycles_t),
						   1,
						   RELAY_SCHEME_LOCKING,
						   NULL,
						   0, 0, 0, 0, 0,
						   0444,
						   NULL,
						   0);

	if (relay_id < 0)
		return -1;
	if (relay_info(relay_id, &chan))
		return -1;
	buf=(cycles_t*)chan.buf_addr;

	if (!buf)
		return -1;
	create_proc_read_entry ( "par_logger",
							 0,
							 NULL,
							 read_proc,
							 (void *)&read_proc_buf_pos );

	if (adeos_register_domain(&domain,&attr)) {
		printk ("par_logger: Cannot register domain (ADEOS).\n");
		return -1;
	}

	printk ("par_logger: Done initializing ...\n");
	return 0;
}

static void __exit clean (void)
{
	printk ("par_logger: nbintrx  %llu\n", nbintrx);
	printk ("par_logger: nbinttx  %llu\n", nbinttx);
	adeos_unregister_domain (&domain);
	relay_close(relay_id);
	remove_proc_entry ("par_logger", NULL);
}

module_init (initialize);
module_exit (clean);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kristian Benoit <kbenoit@opersys.com>");
